import platform as _platform
import re as _re
from datetime import (
    datetime as _datetime,
    timedelta as _timedelta,
    timezone as _timezone,
)
from typing import (
    Optional as _Optional
)


def get_sexagecimal(secs: float, /, include_ms: bool = False) -> str:
    """
    Converts seconds to sexagesimal format.

    ## Demo
    >>> get_sexagecimal(3661.345)
    '01:01:01'

    """
    sign = '-' if secs < 0 else ''
    secs_abs = abs(secs)
    hours, remainder = divmod(secs_abs, 3600)
    minutes, seconds = divmod(remainder, 60)

    h = str(int(hours)).zfill(2)
    m = str(int(minutes)).zfill(2)

    if include_ms:
        s = f'{seconds:.3f}'.zfill(6)
    else:
        s = str(round(seconds)).zfill(2)

    return sign + ':'.join([h, m, s])


def sexagecimal_to_secs(sexagecimal: str, /) -> float:
    """
    ## Exceptions
    - `ValueError` if `sexagecimal` is invalid

    ## Demo
    - `sexagecimal_to_secs('1.25')` -> `1.25`
    - `sexagecimal_to_secs('01:01.25')` -> `61.25`
    - `sexagecimal_to_secs('1:1:5.25')` -> `3665.25`
    """
    _s = sexagecimal.strip(' ')

    res = _re.match(r'^(?P<sign>\+|-)?(?:(?:(?P<h>\d+):)?(?:(?P<m>[0-5]?\d):))?(?P<s>[0-5]?\d(?:\.\d*)?)$', _s)
    if res is None:
        raise ValueError(f'Invalid sexagecimal: {repr(sexagecimal)}')

    sign = res.group('sign')
    if sign in (None, '+'):
        sign = 1
    else:
        sign = -1

    h = res.group('h')
    if h is None:
        h = 0

    m = res.group('m')
    if m is None:
        m = 0

    s = res.group('s')

    return sign * (int(h)*3600 + int(m)*60 + float(s))


def get_dur(__secs: float, /) -> str:
    """
    Converts a duration in seconds to a string in hours, minutes, and seconds format.

    ## Demo
    >>> get_dur(3600)
    '1 hr'
    >>> get_dur(5400)
    '1 hr 30 mins'
    >>> get_dur(7261)
    '2 hrs 1 min 1 sec'
    """
    
    hours, _r = divmod(__secs, 3600)
    minutes, seconds = divmod(_r, 60)

    hours = int(hours)
    minutes = int(minutes)
    seconds = round(seconds)

    parts = []
    
    if hours > 0:
        if hours == 1:
            parts.append('1 hr')
        else:
            parts.append(f'{hours} hrs')
    
    if minutes > 0:
        if minutes == 1:
            parts.append('1 min')
        else:
            parts.append(f'{minutes} mins')

    if seconds == 0:
        if parts == []:
            parts.append('0 sec')
    elif seconds == 1:
        parts.append('1 sec')
    else:
        parts.append(f'{seconds} secs')

    return ' '.join(parts)


class TimeFmt:
    """
    Time Formats: Various datetime presets

    ### Params
    - `timestamp`: If not specified, the current timestamp will be used.
    - `utc_offset`: Current time, but at the specified `utc_offset` (in hours) for the timezone.
                    `2.5` means UTC+02:30, `0` is UTC+0000, `13` is UTC+13, etc.
                    If `None`, will use local timezone.
    """

    def _get_time(ts, utc_offset, fmt):
        dt = _datetime.now() if ts is None else _datetime.fromtimestamp(ts)
        dt = dt.astimezone() if utc_offset is None else dt.astimezone(_timezone(_timedelta(hours=utc_offset)))
        return dt.strftime(fmt)

    def date(timestamp:_Optional[float]=None, utc_offset:_Optional[float]=None) -> str:
        """`Aug 1, 2023` | See class docstring for parameter descriptions."""
        if _platform.system() == 'Windows' : fmt = '%b %#d, %Y'
        elif _platform.system() == 'Linux' : fmt = '%b %-d, %Y'
        elif _platform.system() == 'Darwin': fmt = '%b %-d, %Y'  # macOS
        else: raise NotImplementedError
        return TimeFmt._get_time(timestamp, utc_offset, fmt)

    def hour(timestamp:_Optional[float]=None, utc_offset:_Optional[float]=None) -> str:
        """`HH:MM:SS` / `03:02:01` | See class docstring for parameter descriptions."""
        return TimeFmt._get_time(timestamp, utc_offset, '%H:%M:%S')

    def sort(timestamp:_Optional[float]=None, utc_offset:_Optional[float]=None) -> str:
        """`YYYYMMDD_HHMMSS` / `20231221_013030` | See class docstring for parameter descriptions."""
        return TimeFmt._get_time(timestamp, utc_offset, '%Y%m%d_%H%M%S')

    def neat(timestamp:_Optional[float]=None, utc_offset:_Optional[float]=None) -> str:
        """`2020 Jan 01, 01:02:03 UTC+0000` (fixed length) | See class docstring for parameter descriptions."""
        return TimeFmt._get_time(timestamp, utc_offset, '%Y %b %d, %H:%M:%S UTC%z')

    ## TODO: make tests for this one
    def full(timestamp:_Optional[float]=None, utc_offset:_Optional[float]=None) -> str:
        """`Monday, Jan 1, 2024, 21:12:34 UTC+0000` | See class docstring for parameter descriptions."""
        if _platform.system() == 'Windows' : fmt = '%A, %b %#d, %H:%M:%S UTC%z'
        elif _platform.system() == 'Linux' : fmt = '%A, %b %-d, %H:%M:%S UTC%z'
        elif _platform.system() == 'Darwin': fmt = '%A, %b %-d, %H:%M:%S UTC%z'  # macOS
        else: raise NotImplementedError
        return TimeFmt._get_time(timestamp, utc_offset, fmt)



from typing import (
    NoReturn as _NoReturn,
    Union as _Union,
)

from mykit.inventory.builder import builder as _builder
from mykit.kit.cli_parser.single_simple import SingleSimple as _SingleSimple


__version__ = None  # This value will be written during the build process before production.


def lock_version(version:str, /) -> _Union[None, _NoReturn]:
    """Will raise `AssertionError` if the versions don't match"""
    if version != __version__:
        raise AssertionError(f"The `mykit` version {repr(__version__)} doesn't match the expected {repr(version)}.")


def _main():
    
    p = _SingleSimple('mykit', __version__, 'https://github.com/nvfp/mykit')
    p.add('build', 0)



## dev-docs: This is the variation and the next generation of mykit.kit.path.SafeJSON.
##           SafeJSON will be deprecated soon. TODO: Make NiceJSON, which is the next generation of it.

# import json as _json  # dev-docs: delete this soon since NiceIO is not just for JSON
import os as _os
from typing import (
    Optional as _Optional,
    Tuple as _Tuple,
    List as _List,
    Union as _Union,
)

from mykit.kit.fileops.simple import (
    norm_pth as _norm_pth,
    dont_worry_the_path_ends_with as _dont_worry_the_path_ends_with,
    definitely_a_file as _definitely_a_file,
    definitely_a_dir as _definitely_a_dir,
    this_path_must_not_exist as _this_path_must_not_exist,
)


## birthdate: Oct 1, 2023
class NiceIO:
    """Procedurally strict for performing file operations like read, write, and rewrite"""

    @staticmethod
    def read(file_path:str, /, suffixes:_Optional[_Union[str, _List[str], _Tuple[str, ...]]]=None) -> str:
        """
        Get the content of the given file path

        ## Params
        - `suffixes`: Guarantee reading from the expected file (e.g., making sure reading a Python file by setting `suffixes='.py'`).
                      If `None`, the path's endings will not be checked.

        ## Exceptions
        - Please manually inspect this function for all possible exceptions
        """

        ## Normalize
        path_normalized = _norm_pth(file_path)

        ## Checks
        _definitely_a_file(path_normalized)
        if suffixes is not None: _dont_worry_the_path_ends_with(path_normalized, suffixes)

        ## Read
        with open(path_normalized, 'r') as fp:
            # out = _json.load(fp)  # delete this soon since NiceIO isnt just for JSON
            out = fp.read()

        return out

    @staticmethod
    def write(file_path:str, content:str, suffixes:_Optional[_Union[str, _List[str], _Tuple[str, ...]]]=None) -> None:
        """
        Write the given `content` into a new file

        ## Params
        - `content`: the file contents
        - `suffixes`: Guarantee the expected file (e.g., making sure it's a Python file by setting `suffixes='.py'`)

        ## Exceptions
        - Please manually inspect this function for all possible exceptions
        """

        ## Normalize
        path_normalized = _norm_pth(file_path)

        ## Checks
        if suffixes is not None: _dont_worry_the_path_ends_with(path_normalized, suffixes)
        _definitely_a_dir(_os.path.dirname(path_normalized))
        _this_path_must_not_exist(path_normalized)

        with open(path_normalized, 'w') as fp:
            # _json.dump(content, fp)
            fp.write(content)

    @staticmethod
    def rewrite(file_path:str, new_content:str, suffixes:_Optional[_Union[str, _List[str], _Tuple[str, ...]]]=None) -> None:
        """
        Replace the existing file at `file_path` with the `new_content`

        ## Params
        - `new_content`: the file contents
        - `suffixes`: Guarantee the expected file (e.g., making sure it's a Python file by setting `suffixes='.py'`)

        ## Exceptions
        - Please manually inspect this function for all possible exceptions
        """

        ## Normalize
        path_normalized = _norm_pth(file_path)

        tmp_path = path_normalized + '.tmp'
        bak_path = path_normalized + '.bak'

        ## Checks
        if suffixes is not None: _dont_worry_the_path_ends_with(path_normalized, suffixes)
        _definitely_a_file(path_normalized)
        _this_path_must_not_exist(tmp_path)
        _this_path_must_not_exist(bak_path)

        ## Run
        with open(tmp_path, 'w') as fp: fp.write(new_content)  # writing the new as temp
        _os.rename(path_normalized, bak_path)                  # backup the previous
        _os.rename(tmp_path, path_normalized)                  # rename the temp to the new
        _os.remove(bak_path)                                   # delete the previous

    @staticmethod
    def recover(file_path:str, /, suffixes:_Optional[_Union[str, _List[str], _Tuple[str, ...]]]=None) -> None:
        """
        Try to fix the broken file, mostly because of interrupted or sudden
        stops when performing the `.rewrite` method above. Note that this is still
        not fully optimized and might still change in the future.

        ## Params
        - `suffixes`: Guarantee the expected file (e.g., making sure it's a Python file by setting `suffixes='.py'`)

        ## Exceptions
        - Please manually inspect this function for all possible exceptions
        """

        ## Normalize
        path_normalized = _norm_pth(file_path)

        tmp_path = path_normalized + '.tmp'
        bak_path = path_normalized + '.bak'

        ## Checks
        if suffixes is not None: _dont_worry_the_path_ends_with(path_normalized, suffixes)


        ## case I
        if _os.path.exists(path_normalized) and _os.path.exists(tmp_path) and (not _os.path.exists(bak_path)):
            _os.remove(tmp_path)
            return

        ## case II
        if (not _os.path.exists(path_normalized)) and _os.path.exists(tmp_path) and _os.path.exists(bak_path):
            _os.rename(tmp_path, path_normalized)
            _os.remove(bak_path)
            return

        ## case III
        if _os.path.exists(path_normalized) and (not _os.path.exists(tmp_path)) and _os.path.exists(bak_path):
            _os.remove(bak_path)
            return

        ## case IV (weirdly happening, but currently disabled because i dont know if this is also happening on different environments)
        # if (not _os.path.exists(path_normalized)) and (not _os.path.exists(tmp_path)) and _os.path.exists(bak_path):
        #     _os.rename(bak_path, path_normalized)
        #     return

    @staticmethod
    def erase(file_path:str, /, suffixes:_Optional[_Union[str, _List[str], _Tuple[str, ...]]]=None) -> None:
        """
        ⚠️ Clear the file's content.

        ## Params
        - `suffixes`: Guarantee the expected file (e.g., making sure it's a Python file by setting `suffixes='.py'`)

        ## Exceptions
        - Please manually inspect this function for all possible exceptions
        """
        NiceIO.rewrite(file_path, '', suffixes)



import tkinter as _tk
from typing import (
    Callable as _Callable,
    Dict as _Dict,
    List as _List,
    Literal as _Literal,
    Optional as _Optional,
    Tuple as _Tuple,
    Union as _Union
)

from mykit.app.utils.tagid_manager import TagIdManager as _TagIdManager


class _Rt:  # Runtime

    p: _tk.Canvas = None  # App's main page

    instances = {}  # Identified using ID
    groups    = {}  # Identified using tag


class PhotoImg:

    @staticmethod
    def _install(app):  # For internal use only
        """Fully standalone installation function to plug this component into the app."""
        _Rt.p = app.page

    def __init__(self,
        id: _Optional[str] = None,
        tags: _Optional[_Union[str, _List[str], _Tuple[str, ...]]] = None,
    ) -> None:
        
        ## Handle ID and tags
        _TagIdManager._register_tagid(self, _Rt, id, tags)

    def annihilate(self) -> None:
        """Completely remove it as if it never existed before."""  # dev-docs: use "it" so i dont have to change it for each class
        pass



import random as _random
from typing import (
    Callable as _Callable,
    Dict as _Dict,
    List as _List,
    Literal as _Literal,
    Optional as _Optional,
    Tuple as _Tuple,
    Union as _Union
)

from mykit.app.utils.types import Component as _Component


class TagIdManager:  # TODO: potentially soon this class could be renamed to just 'TagId' if no conflicts occur

    ## Note that a single ID here can now hold multiple components.
    ## This is because the ID is only unique under each component but can be shared overall, to make it more flexible.
    objs_by_id:  _Dict[str, _List[_Component]] = {}  # All components combined
    objs_by_tag: _Dict[str, _List[_Component]] = {}  # All components combined

    @staticmethod
    def get_all_objs_by_id(id:str, /) -> _List[_Component]:
        """
        Get all objects that match the given `id`.

        ## Exceptions
        - `ValueError`: If the given `id` matches no objects.

        ## Docs
        - Note, the returned list is not a copied list, so please be careful when doing something to it.
        """
        if id not in TagIdManager.objs_by_id: raise ValueError(f"There are no objects with id {repr(id)}.")
        return TagIdManager.objs_by_id[id]

    @staticmethod
    def get_obj_by_id(id:str, /) -> _Component:
        """
        Get object with given `id`, believing there is only one object with that ID.

        ## Exceptions
        - `ValueError`: If the given `id` matches no objects.
        - `AssertionError`: If more than one object shares the same `id`.
        """
        if id not in TagIdManager.objs_by_id: raise ValueError(f"There are no objects with id {repr(id)}.")
        objs = TagIdManager.objs_by_id[id]
        if len(objs) != 1: raise AssertionError(f"Multiple objects have the same ID {repr(id)}, when there should be only one.")
        return objs[0]

    @staticmethod
    def get_objs_by_tag(tag:str, /) -> _List[_Component]:
        """
        Get all objects that match the given `tag`.

        ## Exceptions
        - `ValueError`: If the given `tag` matches no objects.

        ## Docs
        - Note, the returned list is not a copied list, so please be careful when doing something to it.
        """
        if tag not in TagIdManager.objs_by_tag: raise ValueError(f"There are no objects with tag {repr(tag)}.")
        return TagIdManager.objs_by_tag[tag]

    @staticmethod
    def _register_tagid(component, runtime, id, tags):  # For internal use only

        ## ID

        if id is None:
            the_id = str(_random.randint(0, 100_000))
            while the_id in runtime.instances:
                the_id = str(_random.randint(0, 100_000))
        else:
            the_id = id
            if the_id in runtime.instances:
                raise ValueError(f'The ID {repr(id)} is duplicated.')

        component.id = the_id  # Assign the ID to the component.
        runtime.instances[the_id] = component  # Register the unique ID in the component ID pool
        if the_id in TagIdManager.objs_by_id:  # Register the ID in the TagIdManager ID pool
            TagIdManager.objs_by_id[the_id].append(component)
        else:
            TagIdManager.objs_by_id[the_id] = [component]

        ## Tags

        if type(tags) is str:
            component.tags = [tags]
        elif (type(tags) is list) or (type(tags) is tuple) or (tags is None):
            component.tags = tags

        if component.tags is not None:
            for tag in component.tags:

                ## Local tags pool (component)
                if tag in runtime.groups: runtime.groups[tag].append(component)
                else: runtime.groups[tag] = [component]

                ## Global tags pool (TagIdManager)
                if tag in TagIdManager.objs_by_tag: TagIdManager.objs_by_tag[tag].append(component)
                else: TagIdManager.objs_by_tag[tag] = [component]



import os as _os
from typing import (
    Optional as _Optional
)


## birthdate: Oct 2, 2023
def make_separator(char:str='─', length:_Optional[int]=None, length_backup:int=21) -> str:
    r"""
    Equivalent to `char*_os.get_terminal_size()[0]`

    ## Params
    - `char`: the separator character (should be a single char)
    - `length`: the separator length
    - `length_backup`: Used with `length=None` when an `OSError` arises (check this function code for details)

    ## Demo
    >>> sep = make_separator('─')
    >>> print(sep + 'hi\n' + sep)
    """

    if length is None:
        try:
            L = _os.get_terminal_size().columns
        except OSError:  # OSError: [Errno 25] Inappropriate ioctl for device
            ## this exception is raised when calling this function inside a VM.
            ## i dont know why, but i guess it's because there's no "physical" terminal,
            ## so there's no sense of terminal dimensions. is this why this exception occurs?
            L = length_backup
    else:
        L = length

    return char*L



import sys as _sys
from typing import (
    Callable as _Callable
)

from mykit.kit.color import (
    Colored as _Colored,
    Hex as _Hex,
)
from mykit.kit.readable.box import box as _box


def _show_help_msg(software, version, repository, cmds):  # dev-docs: the code is a bit messy but compact, it's perfect.
    H = 'show this message then exit'
    longest = max([len(i) for i in cmds.keys()])  # Length of longest command name
    cmds_help = ''
    for name, (_, desc) in cmds.items():
        ## The extra space at the end is one-space padding
        cmds_help += f"   {_Colored(software + ' ' + name, _Hex.CORN)}{' '*(longest-len(name))} -> {_Colored(desc, _Hex.CORN)} \n"
    cmds_help += f"   {_Colored(software, _Hex.CORN)} {' '*longest} -> {_Colored(H, _Hex.CORN)} \n"
    msg = _box(
        '\n'
        ' Commands: \n'  # The extra space at the end is one-space padding (just in case `cmds_help` length < this line length)
        f'{cmds_help}'
        '\n'
        f' Info:\n'
        f'   software  : {software} \n'  # The extra space at the end is one-space padding
        f'   version   : {version} \n'
        f'   repository: {repository} \n'
    )
    print('\n' + msg)


class SingleSimple:
    """
    A simple tool for parsing arguments from the command line, designed to accept just 1 positional argument.

    ## Examples
    >>> myprogram # This will print the help message
    >>> myprogram command_1 # This will run the function associated with command_1
    >>> myprogram command_2 # This will run the function associated with command_2
    """

    def __init__(self, name:str, version:str, repo:str) -> None:
        """
        ## Params
        - `name`: software/program name
        - `version`: software version
        - `repo`: software source code

        ## Docs
        - There is no validation process, such as checking if the command
            name should not be empty, should not be duplicated, etc. So
            please handle that manually.
        - Also, there are no extreme case validations, such as when no
            command was given, etc. Please handle it manually.
        """
        self.name = name
        self.version = version
        self.repo = repo
        self.cmds = {}

    def add(self, cmd:str, run:_Callable[[], None], desc:str) -> None:
        """
        ## Params
        - `cmd`: command name (which will be used like `myprogram <cmd>`)
        - `run`: function to be fired when `cmd` executed
        - `desc`: description of the command
        """
        self.cmds[cmd] = [run, desc]


    def _run_inner(self, args, is_testing):  # Created for testing purposes

        ## Handle show-help
        if len(args) == 1:
            _show_help_msg(self.name, self.version, self.repo, self.cmds)
            if is_testing: return
            _sys.exit(0)

        cmd = args[1]

        ## Handle invalid command
        if cmd not in self.cmds:
            print(f'Unknown commands {repr(cmd)}, run `{self.name}` for help.')
            if is_testing: return
            _sys.exit(1)

        ## Run the corresponding function
        self.cmds[cmd][0]()

    def run(self) -> None:
        """Listen to the command line and run the requested command."""
        args = _sys.argv
        self._run_inner(args, False)



"""
this is the next generation of mykit.kit.color.Colored (mykit-9 ver),
with bugs fixed and even better.
"""
import os as _os
import re as _re
import sys as _sys
from typing import (
    Optional as _Optional,
)

from mykit.kit.color import (
    Hex as _Hex,
    hex_to_rgb as _hex_to_rgb
)
from mykit.kit.str_ops.lcut import lcut as _lcut
from mykit.kit.str_ops.rcut import rcut as _rcut


## The reset code to end the coloring
_RESET = '\033[0m'  # note, "\033" is the same as "\x1b"


class _Runtime:

    ## to make ctxt work inside Windows Command Prompt
    win_init = False


## Birthdate: Oct 6, 2023
def ctxt(text:str, foreground:str=_Hex.LAVENDER_GRAY, background:_Optional[str]=None) -> str:
    """
    Colored Text (ctxt)

    ## Params
    - `text`: The input string, don't worry, `ctxt` will do `text=str(text)` later.
    - `foreground`: in hex format.
    - `background`: in hex format. If `None`, the default terminal background color will be used.

    ## Demo

    ## Docs
    - The text (or parts of the text) that have already been colored cannot have their
        color changed. P.S. I'm not really sure about this one actually.
    """
    
    ## Windows users
    if _sys.platform.lower() == 'win32':
        if not _Runtime.win_init:
            _os.system('color')
            _Runtime.win_init = True

    ## Guard
    text = str(text)

    fg_r, fg_g, fg_b = _hex_to_rgb(foreground)

    header = '\033[' + f'38;2;{fg_r};{fg_g};{fg_b}'

    ## Background
    if background is not None:
        bg_r, bg_g, bg_b = _hex_to_rgb(background)
        header += f';48;2;{bg_r};{bg_g};{bg_b}'
    
    header += 'm'

    """=DOCS=
    color-in-color mechanism
    ========================

    the goals: make it color-in-color properly + no redundancies

    - since all text output from ctxt will have the format H-T-R, where H is the header, T is text, and R is the reset code.

    let's start easy, for a single-color text, it's clear that it's just HTR.

    now, let's add that with plain text, and it becomes HTR_T.

    next, let's add it again with a single-colored text again, and now it becomes HTR_T_HTR.

    now, here is the color-in-color problem. if we just literally put HTR_T_HTR into ctxt, the output would look like H_HTR_T_HTR_R.

    i'm sorry, it's hard to tell which color is which, but let's focus on the order of H and R.

    if we look at H_HTR_T_HTR_R, the middle text (the one that has no color) still has no color (which it's actually should).

    before continuing, here are the things that i observed from a few experiments with it:
    - if there are 2 H's next to each other, the rightmost H will only be used. for example, H1_H2_T, then T will be colored H2.
        also true, if there are 3, 4, or more Hs.
    - all the texts at the right of an H will share the H color. for example, H_T1_T2_T3, all the T1, T2, T3 will have the same color.
    - then, we could say that in order to change/stop the color, we should have either a new H or use R to change or stop the coloring.
    
    
    back again with HTR_T_HTR, one way we could do is H_HTR_HTR_HTR_R. but this is really ridiculous in my mind.
    another way is H_HTRH_T_HTRH_R. here we subtitute all Rs in HTR_T_HTR with RH (R -> RH).

    now, the good news is RH is actually the same as just H. for example, RH_T is the same as H_T.
    also, HR is the same as just R. for example, T_HR is the same as T_R, which is also the same as just T.

    okay, but this is the ideal, HTR_T_HTR -> HT_HT_HTR  (we apply a color to HTR_T_HTR, so it becomes HT_HT_HTR).
    with this version, all H and R are truly doing their job, and there should be no redundancies.

    i have no idea how, but let's list all the ideal cases:
    - HTR_T_HTR -> HT_HT_HTR
    - HTR_T_HTR_T -> HT_HT_HT_HTR
    - T_HTR_T_HTR_T -> HT_HT_HT_HT_HTR
    - H1T_H2TR_T_H2T_H1TR -> H1T_H2T_H3T_H2T_H1TR

    well, it seems there should be only 1 R for each case, which is just at the end.
    or in more proper words, all colored text output from ctxt will just have one R, which is at the end.

    let's back with the simple T, it gonna become HTR, HTR_T is gonna become HT_HTR. T_HTR -> HT_HTR. HTR_T_HTR -> HT_HT_HTR

    okay, hoping that's all

    update: turns out the sequence can't be HT_HT_HT_... because the background is optional.
    if no background at all, it's fine, but if there's a background, the background won't change after the next H.
    the R terminates both foreground and background if two Hs are next to each other, and only the first one has a background.
    the next T will also share the same background. for example, H1T1_H2T2, if H1 has a background but H2 does not,
    T1 and T2 will have the same background (which T2 shouldnt).

    
    so then, the sequence should be HTR_HTR_HTR_ and so on. it's redundant for no BGs, but yeah :)) hopefully soon i could make it better.

    update (final): okay i think im confidently believe that HTR_HTR_HTR... is the safest and ideal sequence for it
    """

    ## <color-in-color>

    ## If the ending is _RESET, remove that so there would be no TR -> TRH -> TRHR (which we expect should be just TR).
    # if text.endswith(_RESET):
    #     text = _rcut(text, _RESET)  # dev-docs: after #1, these lines no longer needed

    ## after #2, it needs these:
    if text.endswith(_RESET):
        text = _rcut(text, _RESET)

    ## Color the uncolored (replace all R with RH)
    # if _RESET in text:
    #     text = text.replace(_RESET, _RESET + header)
    ## vvvvvvvvvvvvvvvvv dev-docs: let's say it's HTR_T_HTR. initially, it becomes HTR_HT_HTR,
    ##                             which is a problem where the last 'T' will have the same background
    ##                             if the middle 'H' has a background while the last 'H' does not.
    ##                             so, it should be HTR_T_HTR -> HT_RHT_RHTR
    # if _RESET in text:
    #     text = text.replace(_RESET, _RESET + header)
    # if '\033[38;2' in text:
    #     text = text.replace(_RESET, _RESET + header)
    ## vvv #1 vvvvvvvvvvvvvvvvv dev-docs: above method will replace all the H and R which is redundant.
    text = _re.sub(r'\033\[0m(?!\033\[38;2)', _RESET+header, text)        # Replace all R (that are not followed by H) with RH
    text = _re.sub(r'(?<!\033\[0m)\033\[38;2', _RESET+'\033[38;2', text)  # Replace all H (that are not preceded by R) with RH


    ## ~~to avoid double coloring (HRH), check if the beginning of `text` is~~
    ## to avoid double coloring (HH), check if the beginning of `text` is
    ## already colored. If it is, no header needs to be inserted at the beginning.
    # if text.startswith(_RESET):
    # if text.startswith('\033['):
    #     header = ''
    ## vvvvvvvvvvvvvvvv dev-docs: after previous step, the begining wil start with _RESET
    # if text.startswith(_RESET):
    #     header = ''  # dev-docs: after #1, these lines no longer needed

    ## after #1, need these:
    # if text.startswith('\033[38;2'):  # starts with H
    #     the_header = ''
    # else:
    #     the_header = header
    # if text.endswith(_RESET):
    #     the_ending = ''
    # else:
    #     the_ending = _RESET
    ## vvv #2 vvvvvvvvv: dev-docs: initially, i thought the final would look like HTR_HTR_HTR, but it turns out to be RHTR_HTR_HTRH.
    ##                             it's fine to just make it HRHTR_HTR_HTRHR, but it's redundant.
    if text.startswith(_RESET):
        text = _lcut(text, _RESET)
        the_header = ''
    else:
        the_header = header

    ## </color-in-color>

    # return _RESET+header + text + _RESET  # should we go with this? im not sure, but i think the R at the end alone should be enough
    # return header + text + _RESET  # dev-docs: after #1, this line is not the right one
    # return text
    ## vvvvv dev-docs: silly, but we back with this again
    # return the_header + text + the_ending
    ## vvv after #2
    return the_header + text + _RESET
